/*
 * Copyright (C) 2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @TODO software trigger test
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <time.h>

#include "wrtd-ut.h"
#include "CuTest.h"

#include <libmockturtle.h>
#include <libwrtd.h>

/* Clear counters. */
static void clear_counters(CuTest *tc, struct wrtd_node *wrtd)
{
        CuAssertIntEquals(tc, 0, wrtd_out_global_counters_reset(wrtd));
        CuAssertIntEquals(tc, 0, wrtd_in_global_counters_reset(wrtd));
        CuAssertIntEquals(tc, 0, wrtd_out_counters_reset(wrtd, 0));
        CuAssertIntEquals(tc, 0, wrtd_in_counters_reset(wrtd, 0));
        CuAssertIntEquals(tc, 0, wrtd_out_counters_reset(wrtd, 1));
        CuAssertIntEquals(tc, 0, wrtd_in_counters_reset(wrtd, 1));
}

static void test_clear_stat(CuTest *tc)
{
	struct wrtd_node *wrtd_0, *wrtd_1;

	wrtd_0 = wrtd_open(2);
	CuAssertPtrNotNull(tc, wrtd_0);
	wrtd_1 = wrtd_open(1);
	CuAssertPtrNotNull(tc, wrtd_1);

        clear_counters(tc, wrtd_0);
        clear_counters(tc, wrtd_1);

	wrtd_close(wrtd_0);
	wrtd_close(wrtd_1);
}

/**
 * For this test there is no special configuration because it uses the
 * internal loopback
 */
static void test_msg_loop(CuTest *tc)
{
	struct wrtd_node *wrtd;
	struct wrtd_trig_id tid = {9, 9, 9};
	struct wrtd_trigger_handle h;
	struct wrtd_input_state sti;
	struct wrtd_output_state sto;

	wrtd = wrtd_open(2);
	CuAssertPtrNotNull(tc, wrtd);

        clear_counters(tc, wrtd);

#if 0
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd), 0));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd), 1));
	sleep(10); /* Wait the White Rabbit link */
#endif

	/* Output */
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd, 0, &h, &tid, NULL));
	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd, &h, 1));

	/* Input */
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd, 0, &tid));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd, 0, 1));
	sleep(5);

	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd, 0, 0));
	sleep(1);
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd, 0, 0));

	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd, 0, &sti));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd, 0, &sto));

	CuAssertIntEquals(tc, sti.total_pulses, sti.sent_triggers);
	CuAssertIntEquals(tc, sti.sent_triggers, sto.received_loopback);
	CuAssertIntEquals(tc, 0, sto.received_messages);
	CuAssertIntEquals(tc, 0, memcmp(&sti.last_sent, &sto.last_received,
					sizeof(struct wrtd_trigger_entry)));

	/* Clear */
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_unassign(wrtd, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd, &h));
	wrtd_close(wrtd);
}

/**
 * For this test you need 2 WRTD boards (LUN 0 LUN 1) and 1 white rabbit switch.
 * You have to connect a pulse generator to the channel 0 of the TDC
 * mezzanine on board with LUN 1
 */
static void test_msg_wr(CuTest *tc)
{
	struct wrtd_node *wrtd_0, *wrtd_1;
	struct wrtd_trig_id tid = {9, 9, 9};
	struct wrtd_trigger_handle h;
	struct wrtd_input_state sti;
	struct wrtd_output_state sto;
	unsigned int prev_lb, prev_msg;

	wrtd_0 = wrtd_open(2);
	CuAssertPtrNotNull(tc, wrtd_0);
	wrtd_1 = wrtd_open(1);
	CuAssertPtrNotNull(tc, wrtd_1);

#if 0
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_0), 0));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_0), 1));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_1), 0));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_1), 1));
	sleep(10); /* Wait the White Rabbit link */
#endif

	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_1, 0, &sto));
	prev_lb = sto.received_loopback;
	prev_msg = sto.received_messages;

	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 0, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_1, 0, 0));

	CuAssertIntEquals(tc, 0, wrtd_in_counters_reset(wrtd_0, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_counters_reset(wrtd_1, 0));

	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_1, 0, &sto));
	prev_lb = sto.received_loopback;
	prev_msg = sto.received_messages;

	/* Input */
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_0, 0, &tid));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_0, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 0, 1));

	/* Output */
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd_1, 0, &h, &tid, NULL));
	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd_1, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_1, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_1, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd_1, &h, 1));

	sleep(5);
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 0, 0));
	sleep(1);
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_1, 0, 0));

	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_0, 0, &sti));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_1, 0, &sto));

	CuAssertIntEquals(tc, sti.sent_triggers,
			  sto.received_messages - prev_msg);
	CuAssertIntEquals(tc, prev_lb, sto.received_loopback);
	CuAssertIntEquals(tc, 0, memcmp(&sti.last_sent, &sto.last_received,
					sizeof(struct wrtd_trigger_entry)));

	/* Clear */
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_0, 0, 0));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_unassign(wrtd_0, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_1, 0, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd_1, &h));
	wrtd_close(wrtd_0);
	wrtd_close(wrtd_1);
}


/**
 * For this test you need 2 WRTD boards (LUN 0 LUN 1) and 1 white rabbit switch.
 * You have to connect a pulse generator to the channel 0 of the TDC
 * mezzanine on board with LUN 1.
 * You have to put the following lemo cables:
 * - LUN 0, FineDelay channel 0 (Out1) -----> LUN 1 TDC channel 1 (In2)
 * - LUN 0, FineDelay channel 1 (Out2) -----> LUN 0 TDC channel 0 (In1)
 * - LUN 1, FineDelay channel 0 (Out1) -----> LUN 0 TDC channel 1 (In2)
 * Path:
 * Lun1/In1 -> tid3 -> Lun0/Out2 ->
 * Lun0/In1 -> tid1 -> Lun1/Out1 ->
 * Lun0/In2 -> tid2 -> Lun0/Out1 -> (loopback)
 * Lun1/In2
 */
static void test_msg_mix(CuTest *tc)
{
	struct wrtd_node *wrtd_0, *wrtd_1;
	struct wrtd_trig_id tid_1 = {0, 0, 1};
	struct wrtd_trig_id tid_2 = {0, 1, 1};
	struct wrtd_trig_id tid_3 = {1, 0, 1};
	struct wrtd_trig_id tid_4 = {1, 1, 1};
	struct wrtd_trigger_handle h_1, h_2, h_3;
	struct wrtd_input_state sti_l1c0, sti_l1c1, sti_l0c0, sti_l0c1;
	struct wrtd_output_state sto_l0c0, sto_l0c1, sto_l1c0;
        const uint64_t tdelay = 200 * 1000 * 1000;

	wrtd_0 = wrtd_open(1);
	CuAssertPtrNotNull(tc, wrtd_0);
	wrtd_1 = wrtd_open(2);
	CuAssertPtrNotNull(tc, wrtd_1);

        clear_counters(tc, wrtd_0);
        clear_counters(tc, wrtd_1);

#if 0
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_0), 0));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_0), 1));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_1), 0));
	CuAssertIntEquals(tc, 0, trtl_cpu_restart(wrtd_get_trtl_dev(wrtd_1), 1));
	sleep(10); /* Wait the White Rabbit link */
#endif

	/* Assign trigger input */
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_0, 0, &tid_1));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_0, 1, &tid_2));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_1, 0, &tid_3));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_1, 1, &tid_4));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_0, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_0, 1, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_1, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_1, 1, WRTD_TRIGGER_MODE_AUTO));

	/* Assign trigger output */
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd_0, 0, &h_2, &tid_2, NULL));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd_0, 1, &h_3, &tid_3, NULL));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd_1, 0, &h_1, &tid_1, NULL));

        CuAssertIntEquals(tc, 0, wrtd_out_trig_delay_set(wrtd_1, &h_1, tdelay));
        CuAssertIntEquals(tc, 0, wrtd_out_trig_delay_set(wrtd_0, &h_2, tdelay));
        CuAssertIntEquals(tc, 0, wrtd_out_trig_delay_set(wrtd_0, &h_3, tdelay));

	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd_0, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd_1, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd_0, 1, WRTD_TRIGGER_MODE_AUTO));

	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_0, 0, 80000000));
	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_1, 0, 80000000));
	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_0, 1, 80000000));
	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_1, 1, 80000000));
	/* Enable and Arm */
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_0, 1, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_1, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 1, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_1, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd_1, &h_1, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd_0, &h_3, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd_0, &h_2, 1));

	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_0, 1, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_1, 1, 1));
	/* Last because connected to pulse generator */
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_1, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 1, 1));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 1, 1));
	/* Last because connected to pulse generator */
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 0, 1));

	sleep(5);

	/* Stop getting pulses from pulse-generator */
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 0, 0));
	sleep(2);
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 1, 0));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 0, 0));
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_0, 1, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 0, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 1, 0));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_1, 0, 0));


	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_0, 0, &sti_l0c0));
	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_0, 1, &sti_l0c1));
	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_1, 0, &sti_l1c0));
	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_1, 1, &sti_l1c1));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_0, 0, &sto_l0c0));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_0, 1, &sto_l0c1));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_1, 0, &sto_l1c0));

	/* Sent and received messages are the same */
	CuAssertIntEquals(tc, sti_l1c0.sent_triggers, sto_l0c1.executed_pulses);
	CuAssertIntEquals(tc, sti_l0c0.sent_triggers, sto_l1c0.executed_pulses);
	CuAssertIntEquals(tc, sti_l0c1.sent_triggers, sto_l0c0.executed_pulses);

        /* Same number of sent and received pulses */
	CuAssertIntEquals(tc, sto_l0c1.executed_pulses, sti_l0c0.sent_triggers);
	CuAssertIntEquals(tc, sto_l1c0.executed_pulses, sti_l0c1.sent_triggers);
	CuAssertIntEquals(tc, sto_l0c0.executed_pulses, sti_l1c1.sent_triggers);

	/* Last sent trigger is the last executed */
	CuAssertIntEquals(tc, sti_l1c0.last_sent.seq,
                          sto_l0c1.last_executed.seq);
	CuAssertIntEquals(tc, sti_l0c0.last_sent.seq,
                          sto_l1c0.last_executed.seq);
	CuAssertIntEquals(tc, sti_l0c1.last_sent.seq,
                          sto_l0c0.last_executed.seq);
	CuAssertIntEquals(tc, 0, memcmp(&sti_l1c0.last_sent.id,
					&sto_l0c1.last_executed.id,
					sizeof(struct wrtd_trig_id)));
	CuAssertIntEquals(tc, 0, memcmp(&sti_l0c0.last_sent.id,
					&sto_l1c0.last_executed.id,
					sizeof(struct wrtd_trig_id)));
	CuAssertIntEquals(tc, 0, memcmp(&sti_l0c1.last_sent.id,
					&sto_l0c0.last_executed.id,
					sizeof(struct wrtd_trig_id)));

	CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd_1, &h_1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd_0, &h_3));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd_0, &h_2));

	/* Close */
	wrtd_close(wrtd_0);
	wrtd_close(wrtd_1);
}

/**
 * For this test you need 2 WRTD boards (LUN 0 LUN 1) and 1 white rabbit switch.
 * You have to connect a pulse generator to the channel 0 of the TDC
 * mezzanine on board with LUN 1.
 * You have to put the following lemo cables:
 * - LUN 0, FineDelay channel 0 (Out1) -----> LUN 1 TDC channel 1 (In2)
 * Path:
 * Lun1/In1 -> tid1 -> Lun0/Out1 -> Lun1/In2
 */
static void test_msg_time(CuTest *tc)
{
	struct wrtd_node *wrtd_0, *wrtd_1;
	struct wrtd_trig_id tid_1 = {0, 2, 1};
	struct wrtd_trig_id tid_2 = {0, 2, 2};
	struct wrtd_trigger_handle h_1;
	struct wrtd_input_state sti1, sti2;
	struct wrtd_output_state sto1;
        const uint64_t tdelay = 200 * 1000 * 1000;

	wrtd_0 = wrtd_open(1);
	CuAssertPtrNotNull(tc, wrtd_0);
	wrtd_1 = wrtd_open(2);
	CuAssertPtrNotNull(tc, wrtd_1);

        clear_counters(tc, wrtd_0);
        clear_counters(tc, wrtd_1);

	/* Assign trigger input */
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_1, 0, &tid_1));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_assign(wrtd_1, 1, &tid_2));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_1, 0, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_trigger_mode_set(wrtd_1, 1, WRTD_TRIGGER_MODE_AUTO));
	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_1, 0, 80 * 1000 * 1000));
	CuAssertIntEquals(tc, 0, wrtd_in_dead_time_set(wrtd_1, 1, 80 * 1000 * 1000));

	/* Assign trigger output */
	CuAssertIntEquals(tc, 0, wrtd_out_trig_assign(wrtd_0, 0, &h_1, &tid_1, NULL));
        CuAssertIntEquals(tc, 0, wrtd_out_trig_delay_set(wrtd_0, &h_1, tdelay));
	CuAssertIntEquals(tc, 0, wrtd_out_pulse_width_set(wrtd_0, 0, 10 * 1000 * 1000ULL));
	CuAssertIntEquals(tc, 0, wrtd_out_trigger_mode_set(wrtd_0, 0, WRTD_TRIGGER_MODE_AUTO));

	/* Enable and Arm */
	CuAssertIntEquals(tc, 0, wrtd_out_arm(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 0, 1));
	CuAssertIntEquals(tc, 0, wrtd_out_trig_enable(wrtd_0, &h_1, 1));

	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_1, 1, 1));
	/* Last because connected to pulse generator */
	CuAssertIntEquals(tc, 0, wrtd_in_arm(wrtd_1, 0, 1));

	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 1, 1));
	/* Last because connected to pulse generator */
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 0, 1));

	sleep(2);

	/* Stop getting pulses from pulse-generator */
	CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 0, 0));
	sleep(2);

        //  Disable channel
        CuAssertIntEquals(tc, 0, wrtd_in_enable(wrtd_1, 1, 0));
        CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 0, 0));
        CuAssertIntEquals(tc, 0, wrtd_out_enable(wrtd_0, 1, 0));

        //  Gather stats.
	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_1, 0, &sti1));
	CuAssertIntEquals(tc, 0, wrtd_out_state_get(wrtd_0, 0, &sto1));
	CuAssertIntEquals(tc, 0, wrtd_in_state_get(wrtd_1, 1, &sti2));

	/* Sent and received messages are the same */
	CuAssertIntEquals(tc, sti1.sent_triggers, sto1.executed_pulses);
	CuAssertIntEquals(tc, sti1.sent_triggers, sti2.sent_triggers);

	/* Last sent trigger is the last executed */
	CuAssertIntEquals(tc, sti1.last_sent.seq, sto1.last_executed.seq);
	CuAssertIntEquals(tc, 0, memcmp(&sti1.last_sent.id,
					&sto1.last_executed.id,
					sizeof(struct wrtd_trig_id)));

        CuAssertIntEquals(tc, 0, wrtd_out_trig_unassign(wrtd_0, &h_1));

	/* Close */
	wrtd_close(wrtd_0);
	wrtd_close(wrtd_1);
}

CuSuite *wrtd_ut_op_suite_get(void)
{
	CuSuite *suite = CuSuiteNew();

	SUITE_ADD_TEST(suite, test_clear_stat);
	SUITE_ADD_TEST(suite, test_msg_loop);
	SUITE_ADD_TEST(suite, test_msg_wr);
	SUITE_ADD_TEST(suite, test_msg_mix);
	SUITE_ADD_TEST(suite, test_msg_time);

	return suite;
}
