/*
 * Copyright (C) 2014-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>

#include "wrtd-ut.h"
#include "CuTest.h"

CuSuite* suite;

static void BuildSuite(void)
{
	suite = CuSuiteNew();

	CuSuiteAddSuite(suite, wrtd_ut_in_suite_get());
	CuSuiteAddSuite(suite, wrtd_ut_out_suite_get());
	CuSuiteAddSuite(suite, wrtd_ut_cmm_suite_get());
	CuSuiteAddSuite(suite, wrtd_ut_op_suite_get());
}

static void run(CuSuite *s)
{
	CuString *output = CuStringNew();

	CuSuiteRun(s, 1);
	CuSuiteSummary(s, output);
	CuSuiteDetails(s, output);
	printf("%s\n", output->buffer);
}

int main(int argc, char *argv[])
{
        int i, j;

        BuildSuite();

        if (argc == 1) {
                run(suite);
        }
        else {
                CuSuite *suite2 = CuSuiteNew();
                for (i = 1; i < argc; i++) {
                        /* Find test for name argv[i]. */
                        for (j = 0; j < suite->count; j++)
                                if (strcmp (argv[i], suite->list[j]->name) == 0)
                                        break;
                        if (j == suite->count) {
                                fprintf(stderr, "cannot find test '%s'\n", argv[i]);
                                return 1;
                        }
                        CuSuiteAdd(suite2, suite->list[j]);
                        run(suite2);
                }
        }

	return 0;
}
