#include <mockturtle-rt.h>

#define MT_WR_LINK_READY (MT_CPU_LR_WR_STAT_LINK_OK | MT_CPU_LR_WR_STAT_TIME_OK)

void send_pkt(int rmq)
{
	struct trtl_fw_msg msg;

	// queue full? wait
	while (mq_claim(TRTL_RMQ, rmq) < 0)
	  ;

	mq_map_out_message(TRTL_RMQ, rmq, &msg);

	msg.header->len = 1;
	*(int *)msg.payload = 0x12345678;

	pp_printf ("TX data\n");

	mq_send(TRTL_RMQ, rmq);
}

void handle_rx(int rmq)
{
	struct trtl_fw_msg tmsg;

	while (!mq_poll_in(TRTL_RMQ, 1 << rmq))
	  ;

	mq_map_in_message(TRTL_RMQ, rmq, &tmsg);

	pp_printf ("RX data: 0x%08x\n", *(int *)tmsg.payload);

	mq_discard(TRTL_RMQ, rmq);
}

int main(void)
{
	int cpu = trtl_get_core_id();
	int rmq = 0;

	struct trtl_ep_eth_address bind_addr;

	pp_printf("TEST for: RMQ UDP %s\n", cpu == 1 ? "recv" : "send");

	while ((lr_readl(MT_CPU_LR_REG_WR_STAT) & MT_WR_LINK_READY) != MT_WR_LINK_READY)
	  ;

	pp_printf("WR link up and time valid\n");

	// set up the RMQ Endpoint

	// we operate only with UDP frames
	bind_addr.type = TRTL_EP_FRAME_UDP;

	// destination MAC: we use broadcast
	bind_addr.dst_mac[0] = 0xff;
	bind_addr.dst_mac[1] = 0xff;
	bind_addr.dst_mac[2] = 0xff;
	bind_addr.dst_mac[3] = 0xff;
	bind_addr.dst_mac[4] = 0xff;
	bind_addr.dst_mac[5] = 0xff;

	// destination port
	bind_addr.dst_port = 12345;
	// source port
	bind_addr.src_port = 7777;
	// destination IP: 192.168.90.255 (broadcast)
	bind_addr.dst_ip = 0xC0A85AFF;
	// source IP: 192.168.90.17
	bind_addr.src_ip = 0xC0A85A11;
	bind_addr.ethertype = 0x800; // IPv4
	// RX filter: we want only UDP packets with matching desination port & IP address.
	bind_addr.filter = TRTL_EP_FILTER_UDP | TRTL_EP_FILTER_DST_PORT | TRTL_EP_FILTER_DST_IP;

	bind_addr.filter |= TRTL_EP_FILTER_ENABLE;

	if (cpu == 1)
	  {
	    // configure incoming channel
	    rmq_bind_in(0, TRTL_EP_ETH, &bind_addr);
	  }
	else
	  {
	    // configure outgoing channel
	    rmq_bind_out(0, TRTL_EP_ETH, &bind_addr);
	  }

	if (cpu == 1) {
	  while (1)
	    handle_rx(rmq);
	}
	else {
	  send_pkt(rmq);
	}

	while (1)
	  ;
	return 0;
}
